require 'xlua'
require 'optim'
require 'cunn'
require 'image'
require 'gnuplot'

local c = require 'trepl.colorize'

opt = lapp[[
    --model                 (default 'split_win19_dep9')             model name
    -g, --gpuid             (default 0)                  gpu id
    --feat_dim              (default 64)

    --data_version          (default 'kitti2015')   
    --data_root             (default '/ais/gobi3/datasets/kitti/scene_flow')
    --perm_fn               (default '')

    --model_param           (default '')    weight file
    --bn_meanstd            (default '')

    --saveDir               (default 'debug')   folder for intermediate prediction result
    
    --sub_width             (default 2000)

    --start_id              (default 161)
    --n                     (default 1)

    --disp_range            (default 128)
    --savebin               (default 0)
    
]]
print(opt)

torch.manualSeed(123)
cutorch.setDevice(opt.gpuid+1)

if opt.data_version == 'modelfish2019' then
    nChannel = 3
    left_folder = 'image_2'
    right_folder = 'image_3'
    groundtruth_folder = 'disp_noc_0'
    err_at = 0.5
else
    error('data_version should be modelfish2019')
end

local function load_model()
    require('models/' .. opt.model .. '.lua')
    bottom, top = split_model(nChannel)
    bottom = bottom:cuda()
    top = top:cuda()

    local bottom_param, bottom_grad_param = bottom:getParameters()
    print(string.format('number of parameters: %d', bottom_param:nElement()))
    
    print(c.blue '==>' ..' loading parameters')
    -- load parameters
    local params = torch.load(opt.model_param)
    assert(params:nElement() == bottom_param:nElement(), string.format('%s: %d vs %d', 'loading parameters: dimension mismatch.', params:nElement(), bottom_param:nElement()))
    bottom_param:copy(params)

    if(string.len(opt.bn_meanstd) > 0) then 
        local bn_mean, bn_std = table.unpack(torch.load(opt.bn_meanstd))

        for k,v in pairs(bottom:findModules('nn.SpatialBatchNormalization')) do
            v.running_mean:copy(bn_mean[k])
            v.running_var:copy(bn_std[k])
        end
        bottom:evaluate() -- bn statistics required
        top:evaluate()
    end
end

function my_forward( m, img )    
    local data = img:view(1, img:size(1), img:size(2), img:size(3))
    local tic = torch.tic()
    for i = 1, #m do
        data = m.modules[i]:updateOutput(data)
        -- print('..' .. i)
        if m.modules[i].finput then
            m.modules[i].finput:set()
        end
    end
    print('feat comp tmr.. ' .. torch.toc(tic))
    return data:clone()
end

if string.len(opt.perm_fn) > 0 then
    file_ids = torch.FloatTensor(torch.FloatStorage(opt.perm_fn))
    print('load permutation: ' .. file_ids:size(1))
else
    file_ids = torch.range(0,199)
    print('use default order: 0-199')
end
paths.mkdir(opt.saveDir)


-- load model
print(c.blue '==>' ..' configuring model')
load_model()

local err_sum = 0.0
local num_imgs = opt.n
print('Num imgs: ', num_imgs)

for i = opt.start_id, opt.start_id+opt.n-1 do
    -- local file_id = file_ids[opt.val_id]
    local file_id = file_ids[i]
    print('--- ' ..  i .. ' --- ' .. ' fn: ' .. file_id)

    local l_fn = string.format('%s/%s/%06d_10.png', opt.data_root, left_folder, file_id)
    local r_fn = string.format('%s/%s/%06d_10.png', opt.data_root, right_folder, file_id)
    local groundtruth_fn = string.format('%s/%s/%06d_10.png', opt.data_root, groundtruth_folder, file_id)
    local l_img = image.load(l_fn, nChannel, 'byte'):cuda()
    local r_img = image.load(r_fn, nChannel, 'byte'):cuda()
    local groundtruth = image.load(groundtruth_fn, 1, 'byte'):cuda()
    l_img:add(-l_img:mean()):div(l_img:std())
    r_img:add(-r_img:mean()):div(r_img:std())

    local img_h = l_img:size(2)
    local img_w = l_img:size(3)
    print('image size: ' .. img_h .. ' x ' .. img_w)

    if opt.small == 1 then
        -- return cuda tensor
        left_feat = bottom:forward(l_img:view(1, nChannel, img_h, img_w)):clone()
        right_feat = bottom:forward(r_img:view(1, nChannel, img_h, img_w)):clone()
    else
        left_feat = my_forward(bottom, l_img)
        right_feat = my_forward(bottom, r_img)
    end

    -- feature should be batch_sz x feat_dim x height x width
    -- print('feature dim: ')
    -- print(#left_feat)
    -- print()

    -- clean bottom memory
    -- clear_model()

    local total_loc = opt.disp_range
    unary_vol = torch.CudaTensor(img_h, img_w, total_loc):zero()
    right_unary_vol = torch.CudaTensor(img_h, img_w, total_loc):zero()

    local tic = torch.tic()

    local start_id, end_id = 1, opt.sub_width
    if end_id > img_w then end_id = img_w end

    -- general(use less memory, compute on sub-image and combine later) but potentially slower
    while start_id <= img_w do
        print(start_id, end_id, img_w)

        for loc_idx = 1, total_loc do
            local x_off = -loc_idx+1 -- always <= 0

            if end_id+x_off >= 1 and img_w >= start_id+x_off then
                --print(math.max(start_id,-x_off+1), math.min(end_id, img_w-x_off))
                local l = left_feat[{{}, {}, {}, {math.max(start_id,-x_off+1), math.min(end_id, img_w-x_off)}}]
                local r = right_feat[{{},{}, {}, {math.max(1,x_off+start_id), math.min(img_w,end_id+x_off)}}]

                local tmp = top:forward({l,r})
                unary_vol[{{},{math.max(start_id, -x_off+1), math.min(end_id, img_w-x_off)},loc_idx}]:copy(tmp)
                
                right_unary_vol[{{},{math.max(1,x_off+start_id), math.min(img_w,end_id+x_off)},loc_idx}]:copy(tmp)
                
            end
        end

        start_id = end_id + 1
        end_id = math.min(img_w, end_id+opt.sub_width)
    end
    print('matching time: ' .. torch.toc(tic))

    -- random output
    -- print('random output: ')
    -- print(unary_vol[{200,499,{51,55}}])
    if opt.savebin == 1 then
        torch.DiskFile(('%s/unary_vol_%06d.bin'):format(opt.saveDir, file_id), 'w'):binary():writeFloat(unary_vol:float():storage())
    end

    paths.mkdir(paths.concat(opt.saveDir, 'unary_img'))
    _,pred = unary_vol:max(3)
    pred = pred:view(img_h, img_w) - 1 -- disp range: [0,...,128], indexing starts at 1 so must substract 1
    --image.save(string.format('%s/unary_img/%06d_10.png', opt.saveDir, file_id), pred:byte())

    -- testing
    local pred_new = torch.CudaTensor(pred:size()):copy(pred)
    local pred_bad = torch.CudaTensor()
    local actual = groundtruth:clone()
    --print(actual[{{},150,500}])
    --print(pred_new[{150,500}])
    
    mask = torch.CudaTensor():resizeAs(actual):ne(actual, 0)
    actual:add(-1.0, pred_new)
    actual = actual:abs()
    pred_bad:gt(actual, err_at):cmul(mask)

    local err = pred_bad:sum() / mask:sum()
    err_sum = err_sum + err


end

mean_test_acc = (1.0-err_sum / num_imgs)
print('Test mean accuracy', mean_test_acc, 'Error thresh:', err_at)




